from rest_framework import serializers

from drf_versioned_schemas.serializers import VersionedModelSerializer

from .models import *


class ExampleModelSerializer(VersionedModelSerializer):
    class Meta:
        model = ExampleModel
        exclude = ['is_active']
        version_model = ExampleModelVersion
        version_model_fields_exclude = ['id', 'model', 'is_active']
        version_field_mapping = {'created_at': 'updated_at'}


class ExampleModelVersionSerializer(serializers.ModelSerializer):
    # https://www.django-rest-framework.org/api-guide/relations/#slugrelatedfield
    tags = serializers.SlugRelatedField(many=True, read_only=True, slug_field='tag')

    class Meta:
        model = ExampleModelVersion
        exclude = ['id', 'model', 'is_active']

    def to_internal_value(self, data):
        ret = super().to_internal_value(data)
        ret['tags'] = data['tags']
        return ret

    def create(self, validated_data):
        tags = validated_data.pop('tags')
        instance = super().create(validated_data)
        for tag in tags:
            instance.tags.create(tag=tag)
        return instance


class ExampleModelWithTagSerializer(VersionedModelSerializer):

    class Meta:
        model = ExampleModel
        exclude = ['is_active']
        version_serializer = ExampleModelVersionSerializer
        version_field_mapping = {'created_at': 'updated_at'}



# Generated by Django 3.2.10 on 2022-09-19 08:57

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Price',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='编辑时间')),
                ('price', models.DecimalField(decimal_places=2, max_digits=10, verbose_name='价格')),
            ],
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='商品ID')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='商品名称')),
            ],
        ),
        migrations.CreateModel(
            name='ProductProfile',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='编辑时间')),
                ('title', models.CharField(max_length=128, verbose_name='商品名称')),
                ('description', models.CharField(max_length=128, verbose_name='商品描述')),
            ],
        ),
    ]



import uuid

from django.db import models
from drf_versioned_schemas.models import VersionedModel, ModelVersion


class Course(VersionedModel):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='课程ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='课程名称')


class CourseVersion(ModelVersion):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='版本ID')
    course = models.ForeignKey(Course, related_name='versions', on_delete=models.CASCADE, verbose_name='数据模型')
    version = models.CharField(max_length=64, verbose_name='版本')
    title = models.CharField(max_length=128, verbose_name='标题')
    created_at = models.DateTimeField(verbose_name='版本创建时间')

    class Meta:
        unique_together = ['course', 'version']


class Lecture(VersionedModel):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='课时ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='课时名称')
    course = models.ForeignKey(Course, related_name='lectures', on_delete=models.CASCADE, verbose_name='课程')
    created_at = models.DateTimeField(verbose_name='课时创建时间')


class LectureVersion(ModelVersion):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='课时版本ID')
    course_version = models.ForeignKey(CourseVersion, related_name='lecture_versions', on_delete=models.CASCADE, verbose_name='课程版本')
    lecture = models.ForeignKey(Lecture, related_name='versions', on_delete=models.CASCADE, verbose_name='课时')
    updated_at = models.DateTimeField(default=None, verbose_name='课时版本更新时间')
    title = models.CharField(max_length=128, verbose_name='课时名称')



# Generated by Django 3.2.10 on 2022-09-20 21:25

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Course',
            fields=[
                ('is_active', models.BooleanField(default=True, verbose_name='是否可用')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='课程ID')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='课程名称')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CourseVersion',
            fields=[
                ('is_active', models.BooleanField(default=True, verbose_name='是否可用')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='版本ID')),
                ('version', models.CharField(max_length=64, verbose_name='版本')),
                ('title', models.CharField(max_length=128, verbose_name='标题')),
                ('created_at', models.DateTimeField(verbose_name='版本创建时间')),
                ('course', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='courses.course', verbose_name='数据模型')),
            ],
            options={
                'unique_together': {('course', 'version')},
            },
        ),
        migrations.CreateModel(
            name='Lecture',
            fields=[
                ('is_active', models.BooleanField(default=True, verbose_name='是否可用')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='课时ID')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='课时名称')),
                ('created_at', models.DateTimeField(verbose_name='课时创建时间')),
                ('course', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='lectures', to='courses.course', verbose_name='课程')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='LectureVersion',
            fields=[
                ('is_active', models.BooleanField(default=True, verbose_name='是否可用')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, verbose_name='课时版本ID')),
                ('updated_at', models.DateTimeField(default=None, verbose_name='课时版本更新时间')),
                ('title', models.CharField(max_length=128, verbose_name='课时名称')),
                ('course_version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='lecture_versions', to='courses.courseversion', verbose_name='课程版本')),
                ('lecture', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='courses.lecture', verbose_name='课时')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import json

from django.test import TestCase

from tests.data import TestDataMixin
from tests.models import ExampleModel, ExampleModelVersion, ExampleModelVersionRelatedField
from tests.serializers import ExampleModelSerializer, ExampleModelWithTagSerializer


class VersionedModelSerializerTestCase(TestDataMixin, TestCase):
    serializer_class = ExampleModelSerializer

    def test_to_representation(self):
        data = self.serializer_class().to_representation(self.instance)
        self.assertDictEqual(self.serialized_data, json.loads(json.dumps(data)))

    def test_serialization(self):
        serializer = self.serializer_class(self.instance)
        self.assertDictEqual(self.serialized_data, json.loads(json.dumps(serializer.data)))

    def test_validate(self):
        serializer = self.serializer_class(self.instance, data=self.new_data)
        self.assertTrue(serializer.is_valid(raise_exception=True))

    def test_to_interval_value(self):
        data = self.serializer_class().to_internal_value(self.new_data)
        self.assertOrderedDictEqual(data, self.deserialized_data)

    def test_create(self):
        serializer = self.serializer_class(data=self.new_data)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        serializer.save()
        self.assertTrue(ExampleModel.objects.get(name=self.new_data['name']))
        self.assertTrue(ExampleModelVersion.objects.filter(version=self.new_data['version']).exists())

    def test_update(self):
        serializer = self.serializer_class(self.instance, data=self.new_version_data)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        serializer.save()
        self.assertTrue(ExampleModelVersion.objects.filter(version=self.new_version_data['version']).exists())

    def test_partial_update(self):
        serializer = self.serializer_class(self.instance, data=self.new_version_partial_data, partial=True)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        serializer.save()
        version_instance = ExampleModelVersion.objects.get(version=self.new_version_partial_data['version'])
        self.assertEqual(version_instance.title, 'Python数据分析')


class CustomVersionSerializerTestCase(TestDataMixin, TestCase):
    serializer_class = ExampleModelWithTagSerializer

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.serialized_data.update({'tags': ['python']})
        cls.new_data.update({'tags': ['python']})
        cls.new_version_data.update({'tags': ['python']})
        cls.deserialized_data['versions'][0].update({'tags': ['python']})

    def test_to_representation(self):
        data = self.serializer_class().to_representation(self.instance)
        self.assertDictEqual(self.serialized_data, json.loads(json.dumps(data)))

    def test_serialization(self):
        serializer = self.serializer_class(self.instance)
        self.assertDictEqual(self.serialized_data, json.loads(json.dumps(serializer.data)))

    def test_to_interval_value(self):
        data = self.serializer_class().to_internal_value(self.new_data)
        self.assertOrderedDictEqual(data, self.deserialized_data)

    def test_create_version(self):
        """
        TODO: 待补充测试
        :return: 
        """
        pass

    def test_create(self):
        serializer = self.serializer_class(data=self.new_data)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        serializer.save()
        model = ExampleModel.objects.get(name=self.new_data['name'])
        self.assertTrue(model)
        model_version = ExampleModelVersion.objects.get(model=model, version=self.new_data['version'])
        self.assertTrue(model_version)
        self.assertTrue(ExampleModelVersionRelatedField.objects.filter(model_version=model_version).exists())

    def test_update(self):
        serializer = self.serializer_class(self.instance, data=self.new_version_data)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        serializer.save()
        model = ExampleModel.objects.get(name=self.new_version_data['name'])
        self.assertTrue(model)
        model_version = ExampleModelVersion.objects.get(model=model, version=self.new_version_data['version'])
        self.assertTrue(model_version)
        self.assertTrue(ExampleModelVersionRelatedField.objects.filter(model_version=model_version).exists())

    def test_partial_update(self):
        serializer = self.serializer_class(self.instance, data=self.new_version_partial_data, partial=True)
        self.assertTrue(serializer.is_valid(raise_exception=True))
        serializer.save()
        model = ExampleModel.objects.get(name=self.new_version_partial_data['name'])
        self.assertTrue(model)
        model_version = ExampleModelVersion.objects.get(model=model, version=self.new_version_partial_data['version'])
        self.assertTrue(model_version)
        self.assertTrue(ExampleModelVersionRelatedField.objects.filter(model_version=model_version).exists())



from django.urls import path, include

from rest_framework.routers import SimpleRouter
from drf_versioned_schemas.routers import ModelVersionSimpleRouter

from .views import CourseViewSet


app_name = 'courses'

router = SimpleRouter()
router.register(r'courses', CourseViewSet)

# version_router = ModelVersionSimpleRouter(router, r'courses', lookup='course')
# version_router.register(r'versions', CourseVersionViewSet, basename='versions')

urlpatterns = [
    path(r'', include(router.urls)),
    # path(r'', include(version_router.urls)),
]



from django.test import SimpleTestCase, override_settings
from django.apps import apps


@override_settings(INSTALLED_APPS=['drf_versioned_schemas.apps.DrfVersionedModelsConfig'])
class AppConfigTestCase(SimpleTestCase):
    def test_install_app(self):
        self.assertTrue(apps.is_installed('drf_versioned_schemas'))
        app_config = apps.get_app_config('drf_versioned_schemas')
        self.assertEqual('drf_versioned_schemas', app_config.name)



from rest_framework.viewsets import ModelViewSet, ReadOnlyModelViewSet


class VersionedModelViewSet(ModelViewSet):
    """
    TODO: 优化异常处理和对应状态码
    """
    pass


class ReadOnlyVersionedModelViewSet(ReadOnlyModelViewSet):
    """
    TODO: 优化异常处理和对应状态码
    """
    pass



"""

"""

from drf_versioned_schemas.viewsets import VersionedModelViewSet

from .models import Course
from .serializers import CourseSerializer


class CourseViewSet(VersionedModelViewSet):
    queryset = Course.objects.all()
    serializer_class = CourseSerializer
    lookup_field = 'name'


class LectureViewSet:
    pass



# -*- coding: utf-8 -*-

from rest_framework import serializers


class VersionedModelSerializer(serializers.ModelSerializer):
    """
    版本化模型序列化类

    假设被序列化的模型由`Model`和`ModelVersion`的最新版本组成。
    其中`Model`定义不可变字段，`ModelVersion`定义可变字段。

    ```python
    class CourseSerializer(VersionedModelSerializer):
        class Meta:
            version_model = CourseVersion
            version_model_fields_exclude = ['id', 'is_active', 'course']
            version_field = 'version'
            version_related_name = 'versions'
            version_field_mapping = {
                'created_at': 'updated_at'
            }
    ```
    """

    # ----- 初始化 -----

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # 版本相关Meta设置
        self._init_version_meta()
        # 验证
        # 版本字段冲突抛出异常，让开发者自己处理
        self._validate_duplicate_fields()
        # partial_update定义为上个版本基础上update字段以后更新
        if self.partial:
            self.initial_data = self._get_partial_update_initial_data(self.initial_data)

    def _init_version_meta(self):
        # 常用Meta设置
        self._model_class = self.Meta.model
        self._model_class_name = self._model_class.__name__
        # 版本序列化类
        self._version_serializer_class = getattr(self.Meta, 'version_serializer', None)
        # 版本模型类
        self._version_model_class = getattr(self.Meta, 'version_model', None) or self._version_serializer_class.Meta.model
        # 版本序列化字段
        # TODO: 给版本序列化类时需要修改
        self._version_model_fields_exclude = getattr(self.Meta, 'version_model_fields_exclude', ['id', 'is_active'])
        # 版本标记字段，默认为`version`
        self._version_field = getattr(self.Meta, 'version_field', 'version')
        # 版本关联字段，默认为`versions`
        self._version_related_name = getattr(self.Meta, 'version_related_name', 'versions')
        # 版本外键字段
        self._version_foreign_key_field_name = getattr(self._model_class, self._version_related_name).field.name
        # 序列化行为版本字段匹配表，默认为空
        self._version_field_mapping = self._validate_version_field_mapping(getattr(self.Meta, 'version_field_mapping', {}))
        # 反序列行为版本字段匹配表，默认为空
        self._version_field_mapping_reverse = {value: key for key, value in getattr(self.Meta, 'version_field_mapping', {}).items()}

    def _validate_version_field_mapping(self, version_field_mapping):
        """
        TODO:
          - 待转字段必须存在
          - key和value不能一样，否则没意义，还影响后面的验证逻辑。

        :param version_field_mapping:
        :return:
        """
        return version_field_mapping

    def get_default_version_serializer_class(self):
        return type(f"{self._model_class_name}Serializer", (serializers.ModelSerializer, ), {
            "Meta": type("Meta", (object,), {
                "model": self._version_model_class,
                "exclude": self._version_model_fields_exclude,
            })
        })

    @property
    def version_serializer_class(self):
        return self._version_serializer_class or self.get_default_version_serializer_class()

    def _validate_duplicate_fields(self):
        """
        TODO:
          - bugfix: 结合mapping表。
          - 增加重复字段提示。
        :return:
        """
        # 模型字段
        model_fields = self.get_fields().keys()
        # 模型版本字段
        model_version_fields = self.version_serializer_class().get_fields().keys()
        # 重复字段
        duplicated_fields = set(model_fields) & set(model_version_fields) - set(self._version_field_mapping.keys())
        if duplicated_fields:
            raise serializers.ValidationError("未处理重复字段")

    def _get_partial_update_initial_data(self, initial_data: dict) -> dict:
        data = self.to_representation(self.instance)
        data.update(initial_data)
        return data

    # ----- 序列化 -----

    def to_representation(self, instance) -> dict:
        """
        :param instance:
        :return:
        """
        instance_latest_version = instance.versions.latest(self._version_field)
        ret = super().to_representation(instance)
        ret_version = self.version_serializer_class().to_representation(instance_latest_version)
        # 处理版本字段名称更变，比如版本`created_at`改为`updated_at`
        ret_version = self._replace_version_field(ret_version)
        ret.update(ret_version)
        return ret

    def _replace_version_field(self, data):
        new_data = {}
        for key in data:
            if key in self._version_field_mapping:
                new_data[self._version_field_mapping[key]] = data[key]
            else:
                new_data[key] = data[key]
        return new_data

    # ----- 反序列化 -----

    def to_internal_value(self, data):
        # 主模型
        model_data = super().to_internal_value(data)
        # 版本模型
        version_data = self._filter_version_data(data, model_data)
        version_data = self.version_serializer_class().to_internal_value(version_data)
        model_data[self._version_related_name] = [version_data]
        return model_data

    def _filter_version_data(self, data, model_data):
        version_data = {}
        for key in data:
            if key not in model_data:
                if key in self._version_field_mapping_reverse:
                    version_data[self._version_field_mapping_reverse[key]] = data[key]
                else:
                    version_data[key] = data[key]
        return version_data

    def create_version(self, instance, version_validated_data):
        version_validated_data.update({self._version_foreign_key_field_name: instance})
        serializer = self.version_serializer_class(data=version_validated_data)
        if serializer.is_valid(raise_exception=True):
            version_instance = serializer.create(version_validated_data)
            return version_instance

    def create(self, validated_data):
        version_validated_data = validated_data.pop(self._version_related_name)[0]
        # 创建新模型
        instance = self._model_class.objects.create(**validated_data)
        # 创建新模型版本
        self.create_version(instance, version_validated_data)
        return instance

    def update(self, instance, validated_data):
        """
        更新模型定义为：
          - 使用现有模型不可变字段（TODO：验证模型不变，具体异常方式待定义）
          - 创建模型版本及可变字段

        :param instance:
        :param validated_data:
        :return:
        """
        # 创建新模型版本
        version_validated_data = validated_data.pop(self._version_related_name)[0]
        self.create_version(instance, version_validated_data)
        return instance



from rest_framework import status
from rest_framework.test import APITestCase


class CourseAPITestCase(APITestCase):
    """
    TODO: 固定手动集成测试。
    """
    fixtures = ['courses.json']

    def test_list_courses(self):
        pass

    def test_curd_course(self):
        pass



"""
测试数据
"""
import json
from collections import OrderedDict
from datetime import datetime

from tests.models import ExampleModel


class TestDataMixin:
    fixtures = ['examples.json']

    @classmethod
    def setUpTestData(cls):
        cls.instance = ExampleModel.objects.get(name='data-analytics-with-python')
        cls.serialized_data = {
            'id': "61741d93-8db4-4845-83c9-7e625e445983",
            'name': 'data-analytics-with-python',
            'version': '0.1.0',
            "title": "Python数据分析",
            "created_at": "2022-05-23T00:00:00",
            "updated_at": '2022-05-23T00:00:00'
        }
        cls.new_data = {
            'name': 'test-name',
            "created_at": "2022-05-23T00:00:00",
            'title': '测试标题',
            'version': '0.1.0',
            'updated_at': '2022-05-24T00:00:00',
        }
        cls.deserialized_data = {
            'name': 'test-name',
            "created_at": datetime.fromisoformat("2022-05-23T00:00:00"),
            'versions': [
                {
                    'version': '0.1.0',
                    'title': '测试标题',
                    'created_at': datetime.fromisoformat('2022-05-24T00:00:00')
                }
            ]
        }
        cls.new_version_data = {
            'name': 'data-analytics-with-python',
            "created_at": "2022-05-23T00:00:00",
            'title': '测试标题2',
            'version': '0.2.0',
            'updated_at': "2022-05-25T00:00:00",
        }
        cls.new_version_partial_data = {
            'name': 'data-analytics-with-python',
            'version': '0.2.1',
            'updated_at': "2022-05-26T00:00:00",
        }

    def assertOrderedDictEqual(self, obj1, obj2):
        def convert_ordered_dict_to_dict(ordered_dict_obj):
            def json_serial(value):
                if isinstance(value, datetime):
                    return value.isoformat()
                return str(value)
            return json.loads(json.dumps(ordered_dict_obj, default=json_serial))
        self.assertDictEqual(convert_ordered_dict_to_dict(obj1), convert_ordered_dict_to_dict(obj2))



from django.test import TestCase
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT
from rest_framework.test import APIRequestFactory

from .models import ExampleModel
from .views import ExampleModelViewSet


class VersionedModelViewSetTestCase(TestCase):

    fixtures = ['examples.json']

    def setUp(self):
        self.factory = APIRequestFactory()

    def test_list(self):
        request = self.factory.get('/examples/')
        example_list_view = ExampleModelViewSet.as_view({'get': 'list'})
        response = example_list_view(request)
        self.assertEqual(response.status_code, HTTP_200_OK)

    def test_retrieve(self):
        request = self.factory.get('/examples/')
        example_detail_view = ExampleModelViewSet.as_view({'get': 'retrieve'})
        response = example_detail_view(request, name='data-analytics-with-python')
        self.assertEqual(response.status_code, HTTP_200_OK)
        self.assertEqual(response.data['name'], 'data-analytics-with-python')
        self.assertEqual(response.data['version'], '0.1.0')
        self.assertEqual(response.data['updated_at'], "2022-05-23T00:00:00")

    def test_create(self):
        data = {
          "name": "cloud-computing-with-python",
          "created_at": "2022-05-23T00:00:00",
          "version": "0.1.0",
          "title": "Python云计算入门",
          "updated_at": "2022-06-25T00:00:00"
        }
        request = self.factory.post('/courses/', data)
        example_list_view = ExampleModelViewSet.as_view({'post': 'create'})
        response = example_list_view(request, name='cloud-computing-with-python')
        self.assertEqual(response.status_code, HTTP_201_CREATED)
        self.assertEqual(response.data['name'], 'cloud-computing-with-python')
        self.assertEqual(response.data['version'], '0.1.0')
        self.assertEqual(response.data['created_at'], "2022-05-23T00:00:00")
        self.assertEqual(response.data['updated_at'], "2022-06-25T00:00:00")

    def test_update(self):
        data = {
          "name": "data-analytics-with-python",
          "created_at": "2022-05-23T00:00:00",
          "version": "0.2.0",
          "title": "Python数据分析",
          "updated_at": "2022-07-25T00:00:00"
        }
        request = self.factory.put('/courses/', data)
        example_detail_view = ExampleModelViewSet.as_view({'put': 'update'})
        response = example_detail_view(request, name='data-analytics-with-python')
        self.assertEqual(response.status_code, HTTP_200_OK)
        self.assertEqual(response.data['name'], 'data-analytics-with-python')
        self.assertEqual(response.data['version'], '0.2.0')
        self.assertEqual(response.data['title'], "Python数据分析")
        self.assertEqual(response.data['created_at'], "2022-05-23T00:00:00")
        self.assertEqual(response.data['updated_at'], "2022-07-25T00:00:00")

    def test_partial_update(self):
        data = {
          "name": "data-analytics-with-python",
          "version": "0.2.1",
          "updated_at": "2022-07-26T00:00:00"
        }
        request = self.factory.patch('/courses/', data)
        example_detail_view = ExampleModelViewSet.as_view({'patch': 'partial_update'})
        response = example_detail_view(request, name='data-analytics-with-python')
        self.assertEqual(response.status_code, HTTP_200_OK)
        self.assertEqual(response.data['name'], 'data-analytics-with-python')
        self.assertEqual(response.data['version'], '0.2.1')
        self.assertEqual(response.data['title'], "Python数据分析")
        self.assertEqual(response.data['created_at'], "2022-05-23T00:00:00")
        self.assertEqual(response.data['updated_at'], "2022-07-26T00:00:00")

    def test_delete(self):
        request = self.factory.delete('/courses/')
        example_detail_view = ExampleModelViewSet.as_view({'delete': 'destroy'})
        response = example_detail_view(request, name='data-analytics-with-python')
        self.assertEqual(response.status_code, HTTP_204_NO_CONTENT)
        self.assertFalse(ExampleModel.objects.get(name='data-analytics-with-python').is_active)




"""
测试模拟声明式配置文件
"""

SECRET_KEY = 'fake-key'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    "tests",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3"
    }
}

ROOT_URLCONF = "tests.urls"



"""

"""

import uuid

from django.db import models


# ----- VersionedModel -----

class VersionedModelManager(models.Manager):
    """
    重新定义版本管理模型的增查删改行为。

    TODO：
      - 异步方法、关联模型等各种情况待验证。
    """
    def get(self, *args, **kwargs):
        """
        "获取"行为定义为获取模型主体和**可用**最新版本。

        :param args:
        :param kwargs:
        :return:
        """
        return super().get(*args, **kwargs)

    def create(self, **kwargs):
        """
        "创建"行为定义为创建不可变字段构成的模型主体和可变字段构成的它的版本。
        :return:
        """
        # 创建模型主体
        model_instance = super().create(**kwargs)
        # 创建模型版本
        model_instance.versions.create(**kwargs)
        return model_instance

    def update(self, **kwargs):
        """
        "更新"行为定义为创建可变字段构成的新版本。

        特殊情况：
        - 如果不可变字段不匹配，则抛出异常；
        - 如果可变字段不全，则拉取上一个最新版本拼接。
        :return:
        """
        # 查找模型
        model_instance = super().get(**kwargs)
        # 查找模型最新版本
        # TODO: 实际上没有get方法
        model_version_latest = model_instance.versions.get()
        # TODO：替换字段
        # 创建模型
        model_instance.versions.create(**kwargs)
        return super().update(**kwargs)


class VersionedModel(models.Model):
    """
    >>> class MyVersionedModel(VersionedModel):
    >>>     pass
    >>>
    >>> class MyModelVersion(models.Model):
    >>>     model = models.ForeignKey(MyVersionedModel, related_name='versions', on_delete=models.CASCADE, verbose_name='数据模型')

    TODO:
      - 允许直接访问ModelVersion的Field。
        比如`course.title`代替`courses.versions.latest('version').title`。
    """
    is_active = models.BooleanField(default=True, verbose_name='是否可用')

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        # TODO: 验证必填Meta
        super(VersionedModel, self).__init__(*args, **kwargs)

    def delete(self) -> None:
        """
        "删除"行为定义为`is_active`标记为`False`。

        ref:
          - https://docs.djangoproject.com/zh-hans/4.1/ref/models/instances/#deleting-objects
          - https://docs.djangoproject.com/zh-hans/4.1/topics/db/queries/#deleting-objects
          - https://docs.djangoproject.com/zh-hans/4.1/topics/db/models/#overriding-predefined-model-methods

        TODO: 返回值未设计
        """
        self.is_active = False
        super().save()


# ----- ModelVersion -----

class ModelVersionManager(models.Manager):
    def update(self):
        # TODO：临时写一下，后面优化。
        raise Exception('禁用')


class ModelVersion(models.Model):
    is_active = models.BooleanField(default=True, verbose_name='是否可用')

    class Meta:
        abstract = True

    class VersionMeta:
        """
        同上。
        """
        # 版本号字段
        version_field = 'version'
        # TODO：一组字段构成的版本号。暂不实现主要是考虑RESTful API不好定义。
        version_fields = ('version', )

    # objects = ModelVersionManager()

    def __init__(self, *args, **kwargs):
        # TODO: 验证必填Meta
        super().__init__(*args, **kwargs)

    def delete(self) -> None:
        """
        "删除"行为定义为`is_active`标记为`False`。

        ref:
          - https://docs.djangoproject.com/zh-hans/4.1/ref/models/instances/#deleting-objects
          - https://docs.djangoproject.com/zh-hans/4.1/topics/db/queries/#deleting-objects

        TODO: 返回值未设计
        """
        self.is_active = False
        super().save()



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys


# import project path to `PYTHONPATH` so that can be found by the Django project,
# but NOT the path of `drf_admin` package!
#
# Ref:
#   - https://stackoverflow.com/questions/3108285/in-python-script-how-do-i-set-pythonpath/3108301

project_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
try:
    sys.path.index(project_path)
except ValueError:
    sys.path.append(project_path)


def main():
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'example.settings')
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()




import uuid

from django.db import models


class Product(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='商品ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='商品名称')


class Price(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    updated_at = models.DateTimeField(auto_now=True, verbose_name='编辑时间')
    price = models.DecimalField(max_digits=10, decimal_places=2, verbose_name='价格')


class ProductProfile(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    updated_at = models.DateTimeField(auto_now=True, verbose_name='编辑时间')
    title = models.CharField(max_length=128, verbose_name='商品名称')
    description = models.CharField(max_length=128, verbose_name='商品描述')



from django.test import TestCase
from django.urls import include, path, reverse, resolve

from rest_framework.test import URLPatternsTestCase
from rest_framework.status import HTTP_200_OK


class VersionedModelRouterTestCase(TestCase, URLPatternsTestCase):
    fixtures = ['examples.json']

    urlpatterns = [
        path('', include('tests.urls'))
    ]

    def test_reverse_example_model_list_url(self):
        url = reverse('tests:examplemodel-list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, HTTP_200_OK)

    def test_resolve_example_model_list_url(self):
        match = resolve('/examples/')
        self.assertEqual(match.view_name, 'tests:examplemodel-list')

    def test_reverse_example_model_detail_url(self):
        url = reverse('tests:examplemodel-detail', kwargs={'name': 'data-analytics-with-python'})
        response = self.client.get(url)
        self.assertEqual(response.status_code, HTTP_200_OK)

    def test_resolve_example_model_detail_url(self):
        match = resolve('/examples/data-analytics-with-python/')
        self.assertEqual(match.view_name, 'tests:examplemodel-detail')



from drf_versioned_schemas.serializers import VersionedModelSerializer

from .models import Course, CourseVersion


class CourseSerializer(VersionedModelSerializer):
    class Meta:
        model = Course
        exclude = ['is_active']
        version_model = CourseVersion
        version_model_fields_exclude = ['id', 'course', 'is_active']
        version_field = 'version'
        version_related_name = 'versions'
        version_field_mapping = {'created_at': 'updated_at'}



from drf_versioned_schemas.viewsets import VersionedModelViewSet

from .models import ExampleModel
from .serializers import ExampleModelSerializer


class ExampleModelViewSet(VersionedModelViewSet):
    queryset = ExampleModel.objects.all()
    serializer_class = ExampleModelSerializer
    lookup_field = 'name'


class LectureViewSet:
    pass



"""
路由类
"""

from rest_framework.routers import SimpleRouter
from rest_framework_nested.routers import NestedSimpleRouter


class ModelVersionMixin:
    def __init__(self, parent_router, parent_prefix, *args, **kwargs):
        pass


class ModelVersionSimpleRouter(ModelVersionMixin, NestedSimpleRouter):
    pass



import unittest

from django.test import TestCase

from drf_versioned_schemas.models import VersionedModel, ModelVersion
from .models import ExampleModel, ExampleModelVersion


class VersionedModelTestCase(TestCase):
    model_class = ExampleModel
    fixtures = ['examples.json']

    def test_init(self):
        model_instance = self.model_class()
        self.assertIsInstance(model_instance, VersionedModel)
        self.assertTrue(model_instance.is_active)

    def test_delete(self):
        model_instance = self.model_class.objects.get()
        model_instance.delete()
        self.assertFalse(model_instance.is_active)


class ModelVersionTestCase(TestCase):
    model_class = ExampleModelVersion
    fixtures = ['examples.json']

    def test_init(self):
        model_instance = self.model_class()
        self.assertIsInstance(model_instance, ModelVersion)
        self.assertTrue(model_instance.is_active)

    def test_delete(self):
        model_instance = self.model_class.objects.get(version='0.1.0')
        model_instance.delete()
        self.assertFalse(model_instance.is_active)




"""
测试数据模型
"""
import uuid

from django.db import models
from drf_versioned_schemas.models import VersionedModel, ModelVersion


# ----- Versioned models -----

class ExampleModel(VersionedModel):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='名称')
    created_at = models.DateTimeField(verbose_name='创建时间')


class ExampleModelVersion(ModelVersion):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='版本ID')
    model = models.ForeignKey(ExampleModel, related_name='versions', on_delete=models.CASCADE, verbose_name='数据模型')
    version = models.CharField(max_length=64, verbose_name='版本')
    title = models.CharField(max_length=128, verbose_name='标题')
    created_at = models.DateTimeField(verbose_name='版本创建时间')

    class Meta:
        unique_together = ['model', 'version']


class ExampleModelVersionRelatedField(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='标签ID')
    model_version = models.ForeignKey(ExampleModelVersion, related_name='tags', on_delete=models.CASCADE, verbose_name='版本模型')
    tag = models.CharField(max_length=16, verbose_name='标签')


# ----- Nested versioned models -----

class ParentModel(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='名称')


class ParentModelVersion(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='版本ID')
    model = models.ForeignKey(ParentModel, related_name='versions', on_delete=models.CASCADE, verbose_name='数据模型')
    title = models.CharField(max_length=128, verbose_name='标题')


class ChildModel(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='名称')
    parent = models.ForeignKey(ParentModel, related_name='children', on_delete=models.CASCADE, verbose_name='父模型')


class ChildModelVersion(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='版本ID')
    parent = models.ForeignKey(ParentModelVersion, related_name='children', on_delete=models.CASCADE, verbose_name='父模型')
    model = models.ForeignKey(ChildModel, related_name='versions', on_delete=models.CASCADE, verbose_name='数据模型')
    title = models.CharField(max_length=128, verbose_name='标题')


# ----- Multi versioned models -----

class MultiVersionedModel(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='ID')
    name = models.CharField(max_length=128, unique=True, verbose_name='名称')


class Part1ModelVersion(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='ID')
    title = models.CharField(max_length=128, verbose_name='标题')


class Part2ModelVersion(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4, verbose_name='ID')
    description = models.CharField(max_length=128, verbose_name='描述')
